#    __    __                   __
#   |  |--|__.-----.-----.-----|__|
#   |     |  |  -__|     |  _  |  |
#   |__|__|__|_____|__|__|_____|__|
#

"""Playground to experiment with 2D particles."""

__title__ = 'hienoi'
__version__ = '0.2.0'
__summary__ = "2D particle playground"
__url__ = 'https://github.com/christophercrouzet/hienoi'
__author__ = "Christopher Crouzet"
__contact__ = 'christopher.crouzet@gmail.com'
__license__ = "MIT"

from hienoi._common import ParticleDisplay
from hienoi._nani import *
from hienoi._numeric import *
from hienoi._vectors import *
